// CHashStr
// http://forlix.org/, df@forlix.org
//
// Copyright (c) 2009-2013 Dominik Friedrichs

#include "CHashStr.h"


CHashStr::CHashStr(unsigned long TableSize)
{
  Size = TableSize;
  // allocate hash table with requested size
  // and clear it
  Table = new HashNd*[Size];
  memset(Table, NULL, sizeof(Table[0])*Size);
}


CHashStr::~CHashStr(void)
{
  // delete all hash nodes
  Wipe();

  // and delete hash table
  delete [] Table;
}


const char *CHashStr::Read(const char *Key,
                           char *Value, size_t ValueLen)
{
  if(!Key || !*Key)
  // null pointer or zero-length key
    return(NULL);

  unsigned long bucket = Bucket(Key, Size);
  HashNd *loopnd = Table[bucket];

  while(loopnd
     && strcmp(loopnd->Key, Key))
  // walk through bucket
  // look for key or end of bucket
    loopnd = loopnd->Link;

  if(!loopnd)
  // key not found
  {
    if(Value && ValueLen)
    // buffer & length provided
      Value[0] = '\0';

    return(NULL);
  }

  if(Value && ValueLen)
  // buffer & length provided
  {
    strncpy(Value, loopnd->Value, ValueLen-1);
    Value[ValueLen-1] = '\0';
  }

  return(loopnd->Value);
}


const char *CHashStr::Read(const char *Key)
{
  // wrapper for quick value access
  return(Read(Key, NULL, NULL));
}


bool CHashStr::Write(const char *Key,
                     const char *Value)
{
  if(!Key || !*Key || !Value)
  // null pointers or zero-length key
    return(false);

  HashNd *newnd = new HashNd;

  // allocate memory for key and value
  newnd->Key = new char[strlen(Key)+1];
  newnd->Value = new char[strlen(Value)+1];

  // fill new node
  strcpy(newnd->Key, Key);
  strcpy(newnd->Value, Value);
  newnd->Link = NULL;

  unsigned long bucket = Bucket(Key, Size);
  HashNd *lastnd = NULL;
  HashNd *loopnd = Table[bucket];

  while(loopnd
     && strcmp(loopnd->Key, Key))
  // walk through bucket, look for key
  // or the end to append
  {
    lastnd = loopnd;
    loopnd = loopnd->Link;
  }

  if(loopnd)
  // key already exists,
  // replace it
  {
    newnd->Link = loopnd->Link;

    delete [] loopnd->Key;
    delete [] loopnd->Value;
    delete loopnd;
  }

  if(!lastnd)
  // first node
    Table[bucket] = newnd;
  else
  // last node or inbetween
    lastnd->Link = newnd;

  return(true);
}


bool CHashStr::Delete(const char *Key)
{
  if(!Key || !*Key)
  // null pointer or zero-length key
    return(false);

  unsigned long bucket = Bucket(Key, Size);
  HashNd *lastnd = NULL;
  HashNd *loopnd = Table[bucket];

  while(loopnd
     && strcmp(loopnd->Key, Key))
  // walk through bucket, look for key
  {
    lastnd = loopnd;
    loopnd = loopnd->Link;
  }

  if(!loopnd)
  // key not found
    return(false);

  if(!lastnd)
  // first node
    Table[bucket] = loopnd->Link;
  else
  // last node or inbetween
    lastnd->Link = loopnd->Link;

  // free node memory
  delete [] loopnd->Key;
  delete [] loopnd->Value;
  delete loopnd;

  return(true);
}


bool CHashStr::LoopThroughKeys(char *&Key,
                               char *&Value,
                               unsigned long *BucketIterator,
                               void *&NodeIterator)
{
  for(HashNd *loopnd = (HashNd *)NodeIterator;
      *BucketIterator < Size;
      (*BucketIterator)++)
  // for each bucket
  {
    // skip empty buckets
    if(!Table[*BucketIterator])
      continue;

    if(!loopnd)
      loopnd = Table[*BucketIterator];

    Key = loopnd->Key;
    Value = loopnd->Value;

    NodeIterator = (void *)loopnd->Link;

    if(!NodeIterator)
      (*BucketIterator)++;

    return(true);
  }

  return(false);
}


unsigned long CHashStr::TellCollisions(void)
{
  unsigned long cols = 0;

  for(unsigned long n = 0;
      n < Size;
      n++)
  // for each bucket
  {
    // skip empty buckets
    if(!Table[n])
      continue;

    HashNd *nextnd = Table[n]->Link;

    while(nextnd)
    // count collisions
    {
      cols++;
      nextnd = nextnd->Link;
    }
  }

  return(cols);
}


void CHashStr::Wipe(void)
{
  for(unsigned long n = 0;
      n < Size;
      n++)
  // for each bucket
  while(Table[n])
  // delete all nodes within
  {
    HashNd *nextnd = Table[n]->Link;

    delete [] Table[n]->Key;
    delete [] Table[n]->Value;
    delete Table[n];

    Table[n] = nextnd;
  }

  return;
}


unsigned long CHashStr::Bucket(const char *key,
                               unsigned long tsize)
{
  unsigned long b = 0;

  // calculate bucket index
  // from given key string
  while(*key)
    b = (b<<6)+(b<<16)-b + *key++;

  b %= tsize;
  return(b);
}
