// Forlix SrcDS Manager
// http://forlix.org/, df@forlix.org
//
// Copyright (c) 2010-2013 Dominik Friedrichs

#include "SrcDS_Manager.h"

VOID WINAPI ServiceMain(DWORD dwArgc, LPTSTR *lpszArgv);
VOID WINAPI ServiceHandler(DWORD fdwControl);

HANDLE g_LogMutex = NULL;
HANDLE g_UpdMutex = NULL;

t_GetProcessMemoryInfo GetProcessMemoryInfo = NULL;

static bool LaunchMonitors(MT_ServerData *mtsd[],
                           size_t MaxServers);
static void TerminateMonitors(MT_ServerData *mtsd[],
                              size_t MaxServers);

static SERVICE_TABLE_ENTRY DispatchTable[] = {{SERVICE_NAME, ServiceMain}, {NULL, NULL}};

static SERVICE_STATUS svcstatus;
static SERVICE_STATUS_HANDLE hsvcstatus;

static DWORD svcmainthreadid = 0;


int main(int argc, char *argv[])
{
  WSADATA WSAData;

  if(WSAStartup(MAKEWORD(2,0), &WSAData))
    return(1);

  char modulepath[MAX_PATH];
  char moduledir[MAX_PATH];

  OSVERSIONINFO osver;
  osver.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);

  GetVersionEx(&osver);

  if(osver.dwPlatformId != VER_PLATFORM_WIN32_NT
  || osver.dwMajorVersion < 5)
  // min os requirement not met
  {
    printf("Requires Win2000/XP or greater\n");
    return(1);
  }

  if(!GetModuleFileName(NULL, modulepath, sizeof(modulepath))
  || !GetFileDirectory(moduledir, modulepath)
  || !SetCurrentDirectory(moduledir))
  // failed to set working directory to where the executable lies
    return(1);

  HMODULE psapi = LoadLibrary("psapi.dll");
  GetProcessMemoryInfo = (t_GetProcessMemoryInfo)GetProcAddress(psapi, "GetProcessMemoryInfo");

  if(!GetProcessMemoryInfo)
  // function address not loaded
    return(1);

  char cmd = '\0';

  if(argc == 2
  && argv[1][0] == '-'
  && argv[1][1] > '\x20'
  && argv[1][2] == '\0')
  // get that char in there
    cmd = tolower(argv[1][1]);

  switch(cmd)
  {
    case 's':
      StartServiceCtrlDispatcher(DispatchTable);
    break;

    case 'i':
    {
      char launch_cmd[MAX_PATH+8];

      sprintf(launch_cmd, "\"%s\" -s", modulepath);
      InstallService(SERVICE_NAME,
                     "Forlix SrcDS Manager",
                     "Monitors multiple Source Dedicated Servers simultaneously, provides "
                     "reliable automatic updating and restarts crashed and unresponsive servers.",
                     launch_cmd,
                     "Tcpip\0AFD\0");
    }
    break;

    case 'u':
      UninstallService(SERVICE_NAME);
    break;

    default:
    {
      printf(APPVERSION "\n"
             "\n"
             "http://forlix.org/, df@forlix.org\n"
             "Copyright (c) 2010-2013 Dominik Friedrichs\n"
             "\n"
             "Supported arguments:\n"
             "\n"
             "  -i  Install service\n"
             "  -u  Uninstall service\n"
             "\n"
             "Place this executable in the desired location the service\n"
             "should run from and then install it using the argument -i.\n"
             "\n"
             "Once the service is installed, restart the system or open the\n"
             "Services management console in order to start the service.\n"
             "\n");

      printf("Press any key to continue...");
      fflush(stdin);
      getch();
      putch('\n');
    }
    break;
  }

  FreeLibrary(psapi);
  WSACleanup();

  return(0);
}


VOID WINAPI ServiceMain(DWORD dwArgc, LPTSTR *lpszArgv)
{
  MT_ServerData *mtsd[MAX_SERVERS] = {NULL};

  hsvcstatus = RegisterServiceCtrlHandler(SERVICE_NAME, ServiceHandler);

  if(!hsvcstatus)
    return;

  svcmainthreadid = GetCurrentThreadId();

  svcstatus.dwCurrentState = SERVICE_START_PENDING;
  svcstatus.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
  svcstatus.dwControlsAccepted = SERVICE_ACCEPT_STOP|SERVICE_ACCEPT_SHUTDOWN;
  svcstatus.dwCheckPoint = 0;
  svcstatus.dwWaitHint = 0;
  svcstatus.dwWin32ExitCode = 0;
  svcstatus.dwServiceSpecificExitCode = 0;
  SetServiceStatus(hsvcstatus, &svcstatus);

  try
  {
    g_LogMutex = CreateMutex(NULL, FALSE, "Forlix_SrcDS_Manager/LogMutex");
    g_UpdMutex = CreateMutex(NULL, FALSE, "Forlix_SrcDS_Manager/UpdMutex");

    if(!g_LogMutex
    || !g_UpdMutex)
      throw(-2);

    LogMsg(APPVERSION " started");

    if(!ReadConfig(mtsd, sizeof(mtsd)/sizeof(mtsd[0]))
    || !LaunchMonitors(mtsd, sizeof(mtsd)/sizeof(mtsd[0])))
      throw(-3);

    svcstatus.dwCurrentState = SERVICE_RUNNING;
    SetServiceStatus(hsvcstatus, &svcstatus);

    for(MSG msg;; Sleep(20))
    // servicemain message loop
    {
      if(PeekMessage(&msg, NULL, WM_APP, WM_APP+0xFF, PM_REMOVE)
      && msg.message == WM_THREAD_TERMINATE)
        throw(0);

      // pipe messaging code...
    }
  }
  catch(int result)
  {
    svcstatus.dwCurrentState = SERVICE_STOP_PENDING;
    SetServiceStatus(hsvcstatus, &svcstatus);

    TerminateMonitors(mtsd, sizeof(mtsd)/sizeof(mtsd[0]));
    LogMsg(APPVERSION " stopped\n");

    CloseHandle(g_LogMutex);
    CloseHandle(g_UpdMutex);

    if(result)
      svcstatus.dwWin32ExitCode = result|(1<<29);

    svcstatus.dwServiceSpecificExitCode = result;
    svcstatus.dwCurrentState = SERVICE_STOPPED;
    SetServiceStatus(hsvcstatus, &svcstatus);
  }

  return;
}


VOID WINAPI ServiceHandler(DWORD fdwControl)
{
  switch(fdwControl)
  {
    case SERVICE_CONTROL_STOP:
    case SERVICE_CONTROL_SHUTDOWN:
    // tell servicemain to terminate itself and all monitor threads
      PostThreadMessage(svcmainthreadid, WM_THREAD_TERMINATE, 0, 0);
    break;

    case SERVICE_CONTROL_INTERROGATE:
      // just send current status
    break;

    default:
      if(fdwControl >= 128
      && fdwControl <= 255)
      {
        // custom control code
      }
    break;
  }

  SetServiceStatus(hsvcstatus, &svcstatus);
  return;
}


static bool LaunchMonitors(MT_ServerData *mtsd[],
                           size_t MaxServers)
{
  unsigned int monitorcnt = 0;

  for(unsigned int i = 0;
      i < MaxServers;
      i++)
  if(mtsd[i])
  {
    mtsd[i]->MonThreadHdl = CreateThread(0, 64*1024, MonitorThread,
                                         mtsd[i], 0, &mtsd[i]->MonThreadId);

    if(!mtsd[i]->MonThreadHdl)
    {
      LogMsg("Failed to create monitor thread for \"%s\"", mtsd[i]->FriendlyName);
      return(false);
    }

    monitorcnt++;
  }

  if(!monitorcnt)
  {
    LogMsg("No servers defined");
    return(false);
  }

  return(true);
}


static void TerminateMonitors(MT_ServerData *mtsd[],
                              size_t MaxServers)
{
  HANDLE qthreads[MAX_SERVERS];
  DWORD qcnt = 0;

  for(unsigned int q = 0; q < MaxServers; q++)
  if(mtsd[q]
  && mtsd[q]->MonThreadHdl)
  // tell all threads to terminate their server and themselves
  // and collect their thread handles to be able to wait for them
  {
    PostThreadMessage(mtsd[q]->MonThreadId, WM_THREAD_TERMINATE, 0, 0);
    qthreads[qcnt++] = mtsd[q]->MonThreadHdl;
  }

  // wait for all the threads to return
  WaitForMultipleObjects(qcnt, qthreads, true, INFINITE);

  for(unsigned int d = 0; d < MaxServers; d++)
  if(mtsd[d])
  // delete all server structs
  {
    CloseHandle(mtsd[d]->MonThreadHdl);

    delete mtsd[d];
    mtsd[d] = NULL;
  }

  return;
}
